/* EPS Conduits (C) 2014-2017 Stephane Charette <stephanecharette@gmail.com>
 * $Id: tun_tap_freebsd.cpp 2167 2017-03-01 12:10:04Z stephane $
 * See license.txt.
 */

#ifdef EPS_FREEBSD

#include "tun_tap.hpp"
#include "logging.hpp"
#include <net/if_tap.h>
#include <net/if.h>
#include <fcntl.h>
#include <errno.h>


bool EPS::validate_network_against_existing_routes(const uint32_t &eps_network_in_host_order, const uint32_t &eps_netmask_in_host_order)
{
	/// @todo Need to find a way to get the existing routes in FreeBSD so we can validate the new EPS address.

	/// @return TRUE if there doesn't seem to be a conflict
	return true;
}


void EPS::set_tun_tap_routes(
	const uint32_t		&network_in_network_order	,
	const uint32_t		&netmask_in_network_order	,
	const std::string	&device_name				)
{
	/// @todo routes not needed at this point in time

	return;
}


void EPS::raise_tun_tap_interface(const uint32_t ip_address_in_network_order, std::string &device_name, EPS_FILE_HANDLE_DESCRIPTOR &device_fd)
{
	device_name	= "";
	device_fd	= -1;

	/* Parts of this function was inspired from OpenVPN, since it turns out that tap devices on BSD are different than
	 * Linux, and not quite how the man page describes them.  Specifically, see the following in OpenVPN:
	 *		src/openvpn/tun.c -> TARGET_FREEBSD -> open_tun_generic() and open_tun()
	 */

	device_fd = open("/dev/tap", O_RDWR);
	if (device_fd < 0)
	{
		THROW_RUNTIME_ERROR(errno, "failed to open /dev/tap (did you remember to \"kldload if_tap\"?)");
	}

	struct ifreq ifr_tun;
	memset(&ifr_tun, '\0', sizeof(ifr_tun));
	int err = ioctl(device_fd, TAPGIFNAME, &ifr_tun);	// "TAP Get InterFace Name"
	if (err < 0)
	{
		THROW_RUNTIME_ERROR(errno, "cannot get the name of the new tap interface: err=" + std::to_string(err));
	}
	device_name = ifr_tun.ifr_name;
	LOG_INFO << "TUN/TAP device " << device_name << " has been created";

	#if 0
	// disable the 4-byte header (I believe this is the same as IFF_NO_PI in Linux)
	const int zero = 0;
	ioctl(device_fd, TUNSIFHEAD, &zero, sizeof(zero));

	const int mode = IFF_BROADCAST | IFF_MULTICAST;
	ioctl(device_fd, TUNSIFMODE, &mode, sizeof(mode));
	#endif

	/* On FreeBSD, the old calls to SIOCSIFADDR were apparently deprecated long ago.
	 * Instead, look up how to use: SIOCAIFADDR (note the subtle 1-character change, S->A)
	 *
	 * For additional information:  man 4 netintro
	 */

	const in_addr_t netmask_in_network_order	= inet_addr("255.255.255.0");
	const in_addr_t broadcast_in_network_order	= ip_address_in_network_order | (~netmask_in_network_order);

	struct ifaliasreq req;
	memset(&req, '\0', sizeof(req));
	strncpy(req.ifra_name, device_name.c_str(), sizeof(req.ifra_name));

	struct sockaddr_in *addr = nullptr;

	addr					= reinterpret_cast<sockaddr_in*>(&req.ifra_addr);
	addr->sin_addr.s_addr	= ip_address_in_network_order;
	addr->sin_family		= AF_INET;
	addr->sin_len			= sizeof(req.ifra_addr);
	LOG_TRACE << device_name << " address .... " << inet_ntoa(addr->sin_addr);

	addr					= reinterpret_cast<sockaddr_in*>(&req.ifra_broadaddr);
	addr->sin_addr.s_addr	= broadcast_in_network_order;
	addr->sin_family		= AF_INET;
	addr->sin_len			= sizeof(req.ifra_broadaddr);
	LOG_TRACE << device_name << " broadcast .. " << inet_ntoa(addr->sin_addr);

	addr					= reinterpret_cast<sockaddr_in*>(&req.ifra_mask);
	addr->sin_addr.s_addr	= netmask_in_network_order;
	addr->sin_family		= AF_INET;
	addr->sin_len			= sizeof(req.ifra_mask);
	LOG_TRACE << device_name << " netmask .... " << inet_ntoa(addr->sin_addr);

	// If we throw we wont close the socket...but then we have bigger issues than a
	// leaked socket and the application will shutdown after the throw anyway...!
	int sock = socket(PF_INET, SOCK_STREAM, 0);
	if (sock < 0)
	{
		THROW_RUNTIME_ERROR(errno, "failed to create a new socket to configure the TUN/TAP interface");
	}

	err = ioctl(sock, SIOCAIFADDR, &req);
	if (err < 0)
	{
		THROW_RUNTIME_ERROR(errno, "failed to set the address for the TUN/TAP interface: err=" + std::to_string(err));
	}

	// get and then set the flags to bring the interface up
	ioctl(sock, SIOCGIFFLAGS, &ifr_tun);
	ifr_tun.ifr_flags |= IFF_UP;
	ifr_tun.ifr_flags |= IFF_RUNNING;
	err = ioctl(sock, SIOCSIFFLAGS, &ifr_tun);
	if (err < 0)
	{
		THROW_RUNTIME_ERROR(errno, "failed to set the flags on the TUN/TAP interface: err=" + std::to_string(err));
	}

	close(sock);

	return;
}


void EPS::close_tun_tap_interface(std::string &device_name, EPS_FILE_HANDLE_DESCRIPTOR &device_fd)
{
	// on FreeBSD, we must explicitely call "ifconfig tapX destroy" otherwise the tap interface remains

	close(device_fd);
	device_fd	= -1;

	if (not device_name.empty())
	{
		struct ifreq ifr_tun;
		memset(&ifr_tun, '\0', sizeof(ifr_tun));
		strcpy(ifr_tun.ifr_name, device_name.c_str());

		int sock	= socket(PF_INET, SOCK_STREAM, 0);
		int err		= ioctl(sock, SIOCIFDESTROY, &ifr_tun);
		LOG_TRACE << "destroying tap device " << ifr_tun.ifr_name << ": err=" << err;
		close(sock);
	}

	return;
}


#endif // EPS_FREEBSD
