/* EPS Conduits (C) 2014-2017 Stephane Charette <stephanecharette@gmail.com>
 * $Id: Hull.hpp 2209 2017-04-26 15:11:53Z stephane $
 * See license.txt.
 */

#pragma once

#include "comm.hpp"
#include "common.hpp"
#include "Tracker.hpp"
#include "NetBuffer.hpp"
#include "Statistics.hpp"
#include "TCPConnection.hpp"
#include <boost/program_options.hpp>
#include <boost/asio/io_service.hpp>
#include <boost/asio/signal_set.hpp>
#include <boost/asio/ip/tcp.hpp>
#include <boost/asio/ip/udp.hpp>
#include <boost/asio/system_timer.hpp>
#include <boost/uuid/uuid.hpp>
#ifndef EPS_WINDOWS
	#include <miniupnpc/miniupnpc.h>
#endif
#include <chrono>
#include <thread>


namespace EPS
{
	/// The @p Hull is the base class for both the client and server. @see EPS::ClientHull @see EPS::ServerHull
	class Hull
	{
		public:

			/// Returns the single instance of Hull, which is at time necessary since the asio service_io is part of the hull.
			static Hull &get(void);

			/// Constructor.
			Hull(const EPS::EMode m, int argc, char **argv);

			/// Destructor.
			virtual ~Hull(void);

			/// Attempt to cleanly shutdown/close all ASIO activity.
			void stop_asio(void);

			/// Start/initialize the signal handling for the hull.
			void start_signal_handling(void);

			/// Get the UUID from the configuration file, or generate a new one if necessary.
			void get_uuid(void);

			/// Save the UUID into the configuration file.
			void save_uuid(void);

			/// Create a new blank configuration file template with helpful comments.
			void write_blank_configuration_file(void);

			/// Read, parse, and store the configuration file.
			void read_configuration_file(void);

			/// Parse command-line options.
			void parse_options(void);

			/// Called by main() to setup exception handling and calls run_second_stage().
			void run_first_stage(void);

			/// Initialize what we need for OpenSSL encryption/decryption.
			void setup_encryption_and_decryption(void);

			/// Pure virtual to run EPS. @see EPS::ClientHull::run_second_stage() @see EPS::ServerHull::run_second_stage()
			virtual void run_second_stage(void) = 0;

			/// Callback when a connection has failed and needs to be cleaned up. @{
			virtual void connection_failed( const EPS::TConnPtr &connection ) = 0;
			virtual void connection_failed( EPS::TCPConnection *connection );
			/// @}

			/// Read (async) a communication command from the peer and act on it.
			void readComm( const EPS::TConnPtr &connection );

			/// Process a communication command that has already been read by readComm().
			virtual void processComm(const EPS::TConnPtr &connection, EPS::NetBuffer *nb);

			/// Derived classes can use this method to get a callback when a communication command is received.
			virtual void processComm(const EPS::TConnPtr &connection, EPS::NetBuffer *nb, const EPS::CommHeader1 *header, const EPS::ECommType type) = 0;

			/// Exchange information and validate connection to peer. @{
			void validate_new_connection( const EPS::TConnPtr &connection );
			void validate_new_connection( EPS::TCPConnection *connection );

			/** Asynchronous write of the buffer.  The de-allocation of the NetBuffer parameter will be handled by
			 * this method once the write has completed.  The NetBuffer must not be modified after this method is
			 * called. @{
			 */
			void async_write( const EPS::TConnPtr &connection, EPS::NetBuffer *nb );
			void async_write( const EPS::TConnPtr &connection );
			/// @}

			/// See if we have peers to which we want to connect.
			void start_connect_timer(void);

			/// Send the latest statistics to the server.
			void send_stats(void);

			/// Start an asynchronous wait on the stats timer.
			void start_stats_timer(void);

			/// Setup socket listener.
			void setup_listener(void);

			/// Make the necessary calls to asynchronously accept a new TCP beam.
			void start_async_accept(void);

			// members

			const EPS::EMode				mode;			///< client or server mode
			const std::string				mode_string;	///< client or server mode string @see EPS::mode_to_string()
			bool							done;			///< %EPS will exit when this is set to @p TRUE
			int								signal_number;	///< set by the signal handler for logging purposes if a fatal signal is raised
			int								ac;				///< argc
			char						**	av;				///< argv
			const std::chrono::seconds		seconds_between_retries;
			EC								ec;				///< boost error code
			EPS::StrMap						configuration;	///< the content of the configuration file
			UUID							uuid;			///< uuid for this %EPS instance

			/// Configuration item read from CLI or eps.conf file. @{
			static std::string				addr;			///< name or ip address to which the client will connect @see @ref cli_server
			static uint16_t					port;			///< TCP/UDP port to which we'll connect @see @ref cli_port
			static std::string				network_str;	///< network (with optional mask) to create for the %EPS conduit network @see @ref cli_network
			static std::string				scorecard;		///< filename of the scorecard where stats are written @see @ref cli_scorecard
			static bool						sequential;		///< should IP addresses be assigned in a sequential manner to new clients @see @ref cli_sequential
			static std::string				conf_filename;	///< name of the configuration file @see @ref cli_conf
			static std::string				log_filename;	///< name of the file to which log entries will be written @see @ref cli_log
			static std::string				gv_filename;	///< name of the graphviz network diagram file @see @ref cli_graphviz
			static bool						upnp_enabled;	///< whether UPnP is enabled/disabled @see @ref cli_upnptimeout
			static int						upnp_timeout;	///< number of seconds to wait for UPnP (or -1 to disable) @see @ref cli_upnptimeout
			/// @}

			boost::asio::io_service			io_service;		///< boost::asio's link to the operating system's I/O services
			boost::asio::signal_set			signal_set;		///< register (through asio) for signal handling
			boost::asio::ip::tcp::resolver	tcp_resolver;	///< tcp resolver for async name resolution
			boost::asio::ip::tcp::acceptor	acceptor;		///< accept incoming client connections

			EPS::IPv4						address;		///< EPS address, e.g. 192.168.43.12
			EPS::IPv4						network;		///< EPS network, e.g. 192.168.43.0
			EPS::IPv4						netmask;		///< EPS netmask, e.g. 255.255.255.0

			EPS::Tracker					tracker;		///< keep track of all clients, uuids, and ip addresses

			EPS::IPv4Set					local_addresses; ///< set of all interesting local addresses

			boost::asio::system_timer		connect_timer;	///< see if we have new connections to establish to create the mesh @see pending_uuids
			boost::asio::system_timer		stats_timer;	///< statistics timer which is triggered every 60 seconds
			EPS::Stats						stats;			///< statistics are updated every minute when @ref stats_timer expires

			IPv4Vec							ipv4_pool;		///< pool of available addresses to give to clients

			EPS::UUIDSet					pending_uuids;	///< set of UUIDs to which we'd like to connect @see connect_timer
			EPS::TCPConns					server_conns;	///< server connections

#ifndef EPS_SKIP_UPNP

			/// UPnP information @{
			boost::asio::system_timer		upnp_renew_timer;	///< timer used to regularly attempt to renew any UPnP port mappings
			EPS::IPv4						upnp_lan_address;	///< LAN ip address according to UPnP
			EPS::IPv4						upnp_wan_address;	///< WAN ip address according to UPnP
			uint16_t						upnp_wan_port;		///< port opened up on WAN side of UPnP device
			UPNPDev						*	upnp_dev;			///< structure used by libminiupnpc
			UPNPUrls						upnp_urls;			///< structure used by libminiupnpc
			IGDdatas						upnp_data;			///< structure used by libminiupnpc
			/// @}

#endif

			time_t							start_time;		///< when EPS was started
	};
}
