/* EPS Conduits (C) 2014-2017 Stephane Charette <stephanecharette@gmail.com>
 * $Id: common.hpp 2208 2017-04-26 07:06:26Z stephane $
 * See license.txt.
 */

#pragma once

#include "eps-windows.hpp"
#include <map>
#include <set>
#include <vector>
#include <string>
#include <boost/asio/ip/tcp.hpp>
#include <boost/asio/ip/address_v4.hpp>
#include <boost/system/error_code.hpp>
#include <boost/uuid/uuid.hpp>

#ifdef EPS_WINDOWS
	#include <boost/asio/windows/stream_handle.hpp>
#else
	#include <boost/asio/posix/stream_descriptor.hpp>
#endif


namespace EPS
{
	// forward declaration
	struct CommData1;

	/// Enum to let some of the common code know if it is working in client or server mode.
	enum EMode
	{
		kInvalid	= 0,
		kClient		= 1,	///< %EPS client mode
		kServer		= 2		///< %EPS server mode
	};

	static const size_t cipher_block_size					= 16;			///< the size of blocks required for encryption/decryption
	static const size_t bps_string_size						= 32;			///< the size of the character-based BPS text fields
	static const size_t network_buffer_size					= 2048;			///< the size of arrays to allocate for incoming, outgoing, encryption and decryption network buffers

	typedef std::map<std::string, std::string>				StrMap;			///< Map of string-to-string.
	typedef boost::asio::ip::tcp::endpoint					TCPEP;			///< Convenient alias for boost TCP endpoint.
	typedef boost::asio::ip::address_v4						IPv4;			///< Convenient alias for boost IPv4 address.
	typedef boost::system::error_code						EC;				///< Convenient alias for boost error codes.
	typedef boost::uuids::uuid								UUID;			///< Convenient alias for boost UUIDs.
	typedef uint8_t											UUID16[16];		///< UUID POD (uuid data)

	typedef std::set<UUID>									UUIDSet;		///< Set of UUIDs
	typedef std::set<EPS::IPv4>								IPv4Set;		///< Set of IPv4 addresses.
	typedef std::vector<EPS::IPv4>							IPv4Vec;		///< Vector of IPv4 addresses.

	#ifdef EPS_WINDOWS
		typedef boost::asio::windows::stream_handle			ASIO_STREAM;				///< TUN/TAP stream handle
		typedef HANDLE										FILE_HANDLE_DESCRIPTOR;		///< native file handle
	#else
		typedef boost::asio::posix::stream_descriptor		ASIO_STREAM;				///< TUN/TAP stream descriptor
		typedef int											FILE_HANDLE_DESCRIPTOR;		///< native file descriptor
	#endif

	/// PKCS padding adds between 1 and 16 bytes to our own length. @{
	size_t number_of_pkcs_padding_bytes( const size_t &len );
	size_t length_including_pkcs_padding( const size_t &len );
	/// @}

	/// Convert a set of IPv4 addresses to a debug string.
	std::string ipset_to_string(const IPv4Set &s);

	/// Convert the mode to a text string.
	std::string mode_to_string(const EPS::EMode mode);

	/// Demangle the given C++ name.  This is compiler-specific.
	std::string cpp_demangle(const std::string &str);

	/// Parse the network+mask string.
	void parse_network_and_mask(const std::string &str, IPv4 &network, IPv4 &netmask);

	/// Convert the protocol field in the TAP packets to a readable string. @{
	std::string proto_to_string(const uint16_t proto_in_host_order);
	std::string proto_to_string(const EPS::CommData1 *	cd1);
	/// @}

	/// Format the value similar to KBytes or KBytes/sec.  @{
	std::string format(const long double &value, std::string &postfix);
	std::string format_bytes(const uint64_t &value);
	std::string format_bps(const uint64_t &value, int seconds);
	/// @}

	/// Get a random int value between the specified range (inclusive).
	int get_random(const int min_val=0, const int max_val=99);

	/// Read the entire file and return it as a text string.
	std::string read_file(const std::string &filename);

	/// Overwrite the filename with this content.
	void write_file(const std::string &filename, const std::string &text);

	/// Produce hex output for the given data.
	std::string hex_out(const uint8_t *data, const size_t len, const int number_of_rows_to_display = 4);

	/// Format the given time (or now) into a UTC timestamp.
	std::string get_utc_timestamp(time_t tt = 0);

	/// Format the given time (or now) into a local timestamp.
	std::string get_local_timestamp(time_t tt = 0);

	/// Format the given bytes to a hex string.  @see EPS::from_hex_string()
	std::string to_hex_string(const uint8_t *data, const size_t len);

	/// Parse the given hex string and populate data.  @see EPS::to_hex_string()
	void from_hex_string(uint8_t *data, const size_t len, std::string str);

	/// Get the content of the template/default eps.conf file.
	std::string get_eps_conf_template(void);

	/// Get a "pretty" text string that describes the OS name.
	std::string get_os_name(void);

	/// Pretty-print a timestamp (in reference to now).
	std::string get_approximate_time(const time_t timestamp);

	/// Return a timestamp representing when the system was last rebooted.
	time_t get_boot_timestamp(void);

	/// Trim whitespace from string. @{
	std::string &trim( std::string &str );
	std::string trim( const std::string &str );
	/// @}
}
