/* EPS Conduits (C) 2014-2017 Stephane Charette <stephanecharette@gmail.com>
 * $Id: comm.cpp 2182 2017-03-26 08:30:05Z stephane $
 * See license.txt.
 */

#include "comm.hpp"
#include "logging.hpp"
#include "common.hpp"
#include <boost/uuid/uuid_io.hpp>
#include <boost/asio.hpp>
#include <limits>
#include <string>
#include <cstring>
#include <time.h>
#ifndef EPS_WINDOWS
	#include <unistd.h>
#endif


static_assert( sizeof(EPS::CommHeader1)								== 16	, "EPS::CommHeader1 should be exactly 16 bytes (one AES block) in length"		);
static_assert( EPS::cipher_block_size								== 16	, "EPS::cipher_block_size should be exactly 16 bytes (one AES block) in length"	);
static_assert( (EPS::network_buffer_size % EPS::cipher_block_size)	== 0	, "EPS::network_buffer_size must be an exact multiple of the cipher block size"	);


EPS::CommHeader1::CommHeader1(void)
{
	memset(this, '\0', sizeof(EPS::CommHeader1));

	return;
}


void EPS::CommHeader1::set(const EPS::ECommType type, const size_t bytes)
{
	// add some randomness to the start of the header block so it doesn't always look the same when we encrypt it
	this->reserved_1		= EPS::get_random(std::numeric_limits<int>::min(), std::numeric_limits<int>::max());
	this->protocol_version	= EPS_PROTOCOL_VERSION;
	this->number_of_bytes	= htons(static_cast<uint16_t>(bytes));
	this->comm_type			=		static_cast<uint8_t >(type	);

	return;
}


size_t EPS::CommHeader1::size_of_packet(void) const
{
	// convert from network to host order, and convert to the expected size_t
	return ntohs(this->number_of_bytes);
}


EPS::ECommType EPS::CommHeader1::type(void) const
{
	return static_cast<EPS::ECommType>(this->comm_type);
}


EPS::CommHello2::CommHello2(void)
{
	memset(this, '\0', sizeof(EPS::CommHello2));

	return;
}


EPS::CommHello2::CommHello2(const UUID &uuid, const EPS::EMode mode, const time_t eps_timestamp, const std::string &cipherhash, const IPv4Set &other_addresses, const IPv4 &address_eps)
{
	this->set(uuid, mode, eps_timestamp, cipherhash, other_addresses, address_eps);

	return;
}


void EPS::CommHello2::set(const UUID &uuid, const EPS::EMode mode, const time_t eps_timestamp, const std::string &cipherhash, const IPv4Set &other_addresses, const IPv4 &address_eps)
{
	memset(this, '\0', sizeof(EPS::CommHello2));

	this->header.set(EPS::kHello2, sizeof(EPS::CommHello2));
	this->mode = static_cast<uint8_t>(mode);

	// first we zero out the address array
	size_t idx = 0;
	for (size_t idx = 0; idx < MAX_NUMBER_OF_OTHER_ADDRESSES; idx++)
	{
		this->other_addresses[idx] = 0;
	}

	// now we store the first few addresses we can fit
	idx = 0;
	for (const auto ip : other_addresses)
	{
		if (idx >= MAX_NUMBER_OF_OTHER_ADDRESSES)
		{
			// we've run out of room and cannot store any more addresses
			break;
		}

		this->other_addresses[idx] = htonl(ip.to_ulong());
		idx ++;
	}

	this->address_eps		= htonl(address_eps.to_ulong());
	this->eps_uptime		= htonl((uint32_t)eps_timestamp);
	this->system_uptime		= htonl((uint32_t)EPS::get_boot_timestamp());

	strncpy(this->uuid_str		, boost::uuids::to_string(uuid).c_str()	, sizeof(uuid_str)		);
	strncpy(this->version_str	, EPS_VERSION							, sizeof(version_str)	);
	strncpy(this->cipherhash_str, cipherhash.c_str()					, sizeof(cipherhash_str));
	LOG_DEBUG << "encryption/decryption cipher hash: " << cipherhash;

	gethostname(this->hostname_str, sizeof(this->hostname_str));
	this->hostname_str[sizeof(this->hostname_str) - 1] = '\0';
	char * ptr = strchr(this->hostname_str, '.');
	if (ptr != nullptr)
	{
		// if the hostname is a fqdn, just take the first part of the name
		ptr[0] = '\0';
	}
	LOG_DEBUG << "hostname: " << this->hostname_str;

	strncpy(this->os_version_str, EPS::get_os_name().c_str(), sizeof(os_version_str));
	this->os_version_str[sizeof(this->os_version_str) - 1] = '\0';

	return;
}


EPS::CommAddress1::CommAddress1(void)
{
	memset(this, '\0', sizeof(EPS::CommAddress1));

	return;
}


EPS::CommAddress1::CommAddress1(const IPv4 &network, const IPv4 &netmask, const IPv4 &address)
{
	this->set(network, netmask, address);

	return;
}


void EPS::CommAddress1::set(const IPv4 &network, const IPv4 &netmask, const IPv4 &address)
{
	memset(this, '\0', sizeof(EPS::CommAddress1));

	this->header.set(EPS::kAddress1, sizeof(EPS::CommAddress1));
	this->network_address	= htonl(network.to_ulong());
	this->network_netmask	= htonl(netmask.to_ulong());
	this->client_address	= htonl(address.to_ulong());

	return;
}


void EPS::CommData1::set(const size_t number_of_data_bytes, const UUID &uuid)
{
	this->header.set(EPS::kData1, number_of_data_bytes + sizeof(EPS::CommData1));

	std::copy(uuid.begin(), uuid.end(), this->dst_uuid);

	return;
}


void EPS::CommList2::set(const size_t number_of_records)
{
	memset(this, '\0', sizeof(EPS::CommList2));

	this->header.set(EPS::kList2, sizeof(EPS::CommList2) + number_of_records * sizeof(EPS::CommList2Data));
	this->number_of_records	= htons( (uint16_t)number_of_records );

	return;
}


void EPS::CommStats1::set(const UUIDSet &uuids)
{
	memset(this, '\0', sizeof(EPS::CommStats1));

	this->header.set(EPS::kStats1, sizeof(EPS::CommStats1) + uuids.size() * sizeof(UUID16));
	this->mesh_connections	= htonl( (uint32_t)uuids.size()	);
	this->timestamp			= htonl( (uint32_t)std::time(nullptr));

	size_t idx = 0;
	for (const auto &u : uuids)
	{
		std::copy(u.begin(), u.end(), this->uuid[idx]);
		idx ++;
	}

	return;
}


std::string EPS::comm_type_to_string( const uint8_t type )
{
	return comm_type_to_string(static_cast<const EPS::ECommType>(type));
}


std::string EPS::comm_type_to_string( const EPS::ECommType type )
{
	return	type == EPS::kUnknown	?	"unknown"	:
			type == EPS::kHello1	?	"hello1"	:
			type == EPS::kAddress1	?	"address1"	:
			type == EPS::kData1		?	"data1"		:
			type == EPS::kList1		?	"list1"		:
			type == EPS::kStats1	?	"stats1"	:
			type == EPS::kHello2	?	"hello2"	:
			type == EPS::kList2		?	"list2"		:
										"invalid (" + std::to_string(type) + ")";
}
